//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

/**
 * \file gtest_factor_inertial_kinematics.cpp
 *
 *  Created on: Janv 29, 2020
 *      \author: Mederic Fourmy
 */

/*


Organisation: For each test, the problem, sensors, factors (if possible) are instanciated in a base class inheriting fromtesting::Test only.
Then, each test case is derived in a child class which defines the data processed by the different sensors and expected estimated test values.
Finally, each of these child classes is used in one or several Test in which basically the expected values are compared against estimation and
solve is done with a perturbated system.

Tests list:

FactorInertialKinematics_1KF_Meas0_bpfix,ZeroMvt:
FactorInertialKinematics_1KF_1p_bpfix,ZeroMvt:
FactorInertialKinematics_1KF_m1p_pfix,ZeroMvt:
FactorInertialKinematics_1KF_1v_bfix,ZeroMvt:
*/


// debug
#include <iostream>
#include <fstream>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>

// WOLF
#include <core/problem/problem.h>
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/math/rotations.h>
#include <core/capture/capture_base.h>
#include <core/feature/feature_base.h>
#include <core/factor/factor_block_absolute.h>
#include <core/state_block/state_block_derived.h>

// Imu
//#include "imu/internal/config.h"
//#include "imu/capture/capture_imu.h"
//#include "imu/processor/processor_imu.h"
//#include "imu/sensor/sensor_imu.h"

// BODYDYNAMICS
#include "bodydynamics/internal/config.h"
#include "bodydynamics/sensor/sensor_inertial_kinematics.h"
#include "bodydynamics/processor/processor_inertial_kinematics.h"
#include "bodydynamics/capture/capture_inertial_kinematics.h"
#include "bodydynamics/feature/feature_inertial_kinematics.h"
#include "bodydynamics/factor/factor_inertial_kinematics.h"
#include <core/state_block/state_block_derived.h>

#include <Eigen/Eigenvalues>

using namespace wolf;
using namespace Eigen;
using namespace std;

const Vector3d ZERO3 = Vector3d::Zero();
const Vector6d ZERO6 = Vector6d::Zero();

class FactorInertialKinematics_1KF : public testing::Test
{
    public:
        wolf::TimeStamp t_;
        ProblemPtr problem_;
        // SensorImuPtr sen_imu;
        SolverCeresPtr solver_;
        // ProcessorBasePtr processor;
        // ProcessorImuPtr processor_imu;
        VectorComposite x_origin_; // initial state
        VectorComposite s_origin_; // initial sigmas for prior
        FrameBasePtr KF0_;
        SensorInertialKinematicsPtr SIK_;
        CaptureInertialKinematicsPtr CIK0_;
        Eigen::Matrix3d Qp_, Qv_, Qw_;
        FeatureInertialKinematicsPtr feat_in_;
        StateBlockPtr sbbimu_;
        Vector3d bias_p_;
        Vector6d bias_imu_;

    void SetUp() override
    {
        //===================================================== SETTING PROBLEM
        // WOLF PROBLEM
        problem_ = Problem::create("POV", 3);

        // CERES WRAPPER
        solver_ = std::make_shared<SolverCeres>(problem_);
        // solver_->getSolverOptions().minimizer_type = ceres::TRUST_REGION; //ceres::TRUST_REGION;ceres::LINE_SEARCH
        // solver_->getSolverOptions().max_line_search_step_contraction = 1e-3;
        // solver_->getSolverOptions().max_num_iterations = 1e4;

        //===================================================== INITIALIZATION
        x_origin_ = VectorComposite("POV", {Vector3d(0,0,0), Quaterniond::Identity().coeffs(), Vector3d(0,0,0)});
        s_origin_ = VectorComposite("POV", {1e-3*Vector3d(1,1,1), 1e-3*Vector3d(1,1,1), 1e-3*Vector3d(1,1,1)});
        t_.set(0.0);
        KF0_ = problem_->setPriorFactor(x_origin_, s_origin_, t_);

        ParamsSensorInertialKinematicsPtr intr_ik = std::make_shared<ParamsSensorInertialKinematics>();
        SIK_ = std::make_shared<SensorInertialKinematics>(Eigen::VectorXd(0), intr_ik);

        // ADD THE STATEBLOCKS AND SET THEIR DEFAULT VALUES
        bias_p_ = ZERO3;
        bias_imu_ = ZERO6;
        StateBlockPtr sbc = make_shared<StatePoint3d>(ZERO3);
        StateBlockPtr sbL = make_shared<StateVector3d>(ZERO3);
        StateBlockPtr sbd = make_shared<StateVector3d>(ZERO3);
        StateBlockPtr sbb = make_shared<StateParams3>(bias_p_);
        sbbimu_ = make_shared<StateParams6>(bias_imu_);

        KF0_->addStateBlock('C', sbc, problem_);
        KF0_->addStateBlock('D', sbd, problem_);
        KF0_->addStateBlock('L', sbL, problem_);
        KF0_->addStateBlock('I', sbbimu_, problem_);  // IMU bias

        // Fix the one we cannot estimate
        KF0_->getStateBlock('I')->fix();  // Imu

        // =============== SET DEFAULT" KIN MEASUREMENT AND COVARIANCE
        // Measurements
        Vector3d pBC_meas = ZERO3;
        Vector3d vBC_meas = ZERO3;
        Vector3d w_meas = ZERO3;

        // momentum parameters
        Matrix3d Iq; Iq.setIdentity();
        Vector3d Lq = ZERO3;

        Qp_ = pow(1e-2, 2)*Eigen::Matrix3d::Identity();
        Qv_ = pow(1e-2, 2)*Eigen::Matrix3d::Identity();
        Qw_ = pow(1e-2, 2)*Eigen::Matrix3d::Identity();

        // =============== CREATE CAPURE
        Vector9d meas_ikin0; meas_ikin0 << pBC_meas, vBC_meas, w_meas;
        CIK0_ = CaptureBase::emplace<CaptureInertialKinematics>(KF0_, t_, SIK_, meas_ikin0, Iq, Lq, bias_p_);
    }

    void TearDown() override{}
};


class FactorInertialKinematics_1KF_Meas0_bpfix : public FactorInertialKinematics_1KF
{
    void SetUp() override
    {
        FactorInertialKinematics_1KF::SetUp();
        // Fix the bp bias
        CIK0_->getStateBlock('I')->fix();
        KF0_->getStateBlock('C')->unfix();
        Eigen::Vector3d ZERO3 = Eigen::Vector3d::Zero();

        Eigen::Vector3d pBC_meas = ZERO3;
        Eigen::Vector3d vBC_meas = ZERO3;
        Eigen::Vector3d w_meas = ZERO3;
        Eigen::Vector9d meas; meas << pBC_meas, vBC_meas, w_meas;
        // momentum parameters
        Eigen::Matrix3d Iq; Iq.setIdentity();
        Eigen::Vector3d Lq = ZERO3;

        // =============== CREATE FEATURE/FACTOR
        Eigen::Matrix9d Q_ikin_err = computeIKinCov(Qp_, Qv_, Qw_, meas.head<3>()-bias_p_, meas.tail<3>()-bias_imu_.tail<3>(), Iq);
        feat_in_ = FeatureBase::emplace<FeatureInertialKinematics>(CIK0_, meas, Q_ikin_err, Iq, Lq, FeatureBase::UncertaintyType::UNCERTAINTY_IS_COVARIANCE);
        auto fac = FactorBase::emplace<FactorInertialKinematics>(feat_in_, feat_in_, sbbimu_, nullptr, false);
    }
};

class FactorInertialKinematics_1KF_1p_bpfix : public FactorInertialKinematics_1KF
{
    void SetUp() override
    {
        FactorInertialKinematics_1KF::SetUp();
        // Fix the bp bias
        CIK0_->getStateBlock('I')->fix();
        KF0_->getStateBlock('C')->unfix();
        Eigen::Vector3d ZERO3 = Eigen::Vector3d::Zero();

        Eigen::Vector3d pBC_meas; pBC_meas << 1,0,0;
        Eigen::Vector3d vBC_meas = ZERO3;
        Eigen::Vector3d w_meas = ZERO3;
        Eigen::Vector9d meas; meas << pBC_meas, vBC_meas, w_meas;
        // momentum parameters
        Eigen::Matrix3d Iq; Iq.setIdentity();
        Eigen::Vector3d Lq = ZERO3;

        // =============== CREATE FEATURE/FACTOR
        Eigen::Matrix9d Q_ikin_err = computeIKinCov(Qp_, Qv_, Qw_, meas.head<3>()-bias_p_, meas.tail<3>()-bias_imu_.tail<3>(), Iq);
        feat_in_ = FeatureBase::emplace<FeatureInertialKinematics>(CIK0_, meas, Q_ikin_err, Iq, Lq, FeatureBase::UncertaintyType::UNCERTAINTY_IS_COVARIANCE);
        auto fac = FactorBase::emplace<FactorInertialKinematics>(feat_in_, feat_in_, sbbimu_, nullptr, false);
    }
};

class FactorInertialKinematics_1KF_m1p_pfix : public FactorInertialKinematics_1KF
{
    void SetUp() override
    {
        FactorInertialKinematics_1KF::SetUp();
        // Fix the COM position
        CIK0_->getStateBlock('I')->unfix();
        KF0_->getStateBlock('C')->fix();
        Eigen::Vector3d ZERO3 = Eigen::Vector3d::Zero();

        Eigen::Vector3d pBC_meas; pBC_meas << 1,0,0;
        Eigen::Vector3d vBC_meas = ZERO3;
        Eigen::Vector3d w_meas = ZERO3;
        bias_p_ << 1,0,0;
        Eigen::Vector9d meas; meas << pBC_meas, vBC_meas, w_meas;
        // momentum parameters
        Eigen::Matrix3d Iq; Iq.setIdentity();
        Eigen::Vector3d Lq = ZERO3;

        // =============== CREATE FEATURE/FACTOR
        Eigen::Matrix9d Q_ikin_err = computeIKinCov(Qp_, Qv_, Qw_, meas.head<3>()-bias_p_, meas.tail<3>()-bias_imu_.tail<3>(), Iq);
        feat_in_ = FeatureBase::emplace<FeatureInertialKinematics>(CIK0_, meas, Q_ikin_err, Iq, Lq, FeatureBase::UncertaintyType::UNCERTAINTY_IS_COVARIANCE);
        auto fac = FactorBase::emplace<FactorInertialKinematics>(feat_in_, feat_in_, sbbimu_, nullptr, false);
    }
};

class FactorInertialKinematics_1KF_1v_bfix : public FactorInertialKinematics_1KF
{
    void SetUp() override
    {
        FactorInertialKinematics_1KF::SetUp();
        // Fix the bp bias
        Eigen::Vector3d ZERO3 = Eigen::Vector3d::Zero();
        CIK0_->getStateBlock('I')->fix();
        KF0_->getStateBlock('C')->unfix();
        bias_p_ = ZERO3;
        CIK0_->getStateBlock('I')->setState(bias_p_);

        Eigen::Vector3d pBC_meas = ZERO3;
        Eigen::Vector3d vBC_meas; vBC_meas << 1,0,0;
        Eigen::Vector3d w_meas = ZERO3;
        Eigen::Vector9d meas; meas << pBC_meas, vBC_meas, w_meas;
        // momentum parameters
        Eigen::Matrix3d Iq; Iq.setIdentity();
        Eigen::Vector3d Lq = ZERO3;

        // =============== CREATE FEATURE/FACTOR
        Eigen::Matrix9d Q_ikin_err = computeIKinCov(Qp_, Qv_, Qw_, meas.head<3>()-bias_p_, meas.tail<3>()-bias_imu_.tail<3>(), Iq);
        feat_in_ = FeatureBase::emplace<FeatureInertialKinematics>(CIK0_, meas, Q_ikin_err, Iq, Lq, FeatureBase::UncertaintyType::UNCERTAINTY_IS_COVARIANCE);
        auto fac = FactorBase::emplace<FactorInertialKinematics>(feat_in_, feat_in_, sbbimu_, nullptr, false);
    }
};


//class FactorInertialKinematic_imu : public testing::Test
//{
//    public:
//        wolf::TimeStamp t_;
//        ProblemPtr problem_;
//        SolverCeres* solver_;
//        VectorXd x_origin_;
//        MatrixXd P_origin_;
//        FrameBasePtr KF0_;
//        Eigen::Matrix3d Qp_, Qv_, Qw_;
//        Eigen::Vector3d bias_p_, bias_imu_;
//        FeatureInertialKinematicsPtr feat_in_;
//        SensorImuPtr sen_imu_;
//        ProcessorImuPtr processor_imu_;
//
//    virtual void SetUp()
//    {
//        // using shared_ptr;
//        // using make_shared;
//        // using static_pointer_cast;
//
//        std::string bodydynamics_root = _WOLF_BODYDYNAMICS_ROOT_DIR;
//
//        //===================================================== SETTING PROBLEM
//        // WOLF PROBLEM
//        problem_ = Problem::create("POV", 3);
//
//        // CERES WRAPPER
//        ceres::Solver::Options ceres_options;
//        // ceres_options.minimizer_type = ceres::TRUST_REGION; //ceres::TRUST_REGION;ceres::LINE_SEARCH
//        // ceres_options.max_line_search_step_contraction = 1e-3;
//        // ceres_options.max_num_iterations = 1e4;
//        solver_ = new SolverCeres(problem_, ceres_options);
//
//        //===================================================== INITIALIZATION
//        x_origin_.resize(10);
//        x_origin_ << 0,0,0, 0,0,0,1, 0,0,0;
//        P_origin_ = pow(1e-3, 2) * Eigen::Matrix9d::Identity();
//        t_.set(0);
//
//
//        // Set origin of the problem
//        KF0_ = problem_->setPrior(x_origin_, P_origin_.topLeftCorner<6,6>(), t_, 0.005);
//        // Specific factor for origin velocity (not covered by setPrior)
//        CaptureBasePtr capV0 = CaptureBase::emplace<CaptureBase>(KF0_, "Vel0", t_);
//        FeatureBasePtr featV0 = FeatureBase::emplace<FeatureBase>(capV0, "Vel0", x_origin_.tail(3), P_origin_.bottomRightCorner<3,3>());
//        FactorBasePtr facV0 = FactorBase::emplace<FactorBlockAbsolute>(featV0, KF0_->getV());
//
//
//        // SENSOR + PROCESSOR Imu
//        SensorBasePtr sen0_ptr = problem_->installSensor("SensorImu", "Main Imu", (Vector7d()<<0,0,0,0,0,0,1).finished(), bodydynamics_root + "/demos/sensor_imu.yaml");
//        sen_imu_ = std::static_pointer_cast<SensorImu>(sen0_ptr);
//        ProcessorBasePtr proc = problem_->installProcessor("ProcessorImu", "Imu pre-integrator", "Main Imu", bodydynamics_root + "/demos/processor_imu.yaml");
//        processor_imu_ = std::static_pointer_cast<ProcessorImu>(proc);
//
//
//        // ADD THE STATEBLOCKS AND SET THEIR DEFAULT VALUES
//        Vector3d ZERO3; ZERO3 << 0,0,0;
//        Vector6d ZERO6; ZERO6 << ZERO3, ZERO3;
//        Vector3d bias_p_ = ZERO3;
//        Vector6d bias_imu_ = ZERO6;
//        StateBlockPtr sbc = make_shared<StatePoint3d>(ZERO3);
//        StateBlockPtr sbd = make_shared<StateBlock>(ZERO3);
//        StateBlockPtr sbL = make_shared<StateBlock>(ZERO3);
//        StateBlockPtr sbb = make_shared<StateBlock>(bias_p_);
//
//        KF0_->addStateBlock("C", sbcproblem_);
//        KF0_->addStateBlock("D", sbdproblem_);
//        KF0_->addStateBlock("L", sbLproblem_);
//
//        // Fix the one we cannot estimate
//        // KF0_->getP()->fix();
//        // KF0_->getO()->fix();
//        // KF0_->getV()->fix();
//
//        // =============== SET DEFAULT" KIN MEASUREMENT AND COVARIANCE
//        // Measurements
//        Vector3d pBC_meas = ZERO3;
//        Vector3d vBC_meas = ZERO3;
//        Vector3d w_meas = ZERO3;
//
//        // momentum parameters
//        Matrix3d Iq; Iq.setIdentity();
//        Vector3d Lq = ZERO3;
//
//        Qp_ = pow(1e-2, 2)*Eigen::Matrix3d::Identity();
//        Qv_ = pow(1e-2, 2)*Eigen::Matrix3d::Identity();
//        Qw_ = pow(1e-2, 2)*Eigen::Matrix3d::Identity();
//        // Qn.setZero();
//        // Qn.topLeftCorner<3,3>() = Qp_;
//        // Qn.block<3,3>(3,3) = Qv_;
//        // Qn.bottomRightCorner<3,3>() = Qw_;
//
//        Matrix9d Qkin = computeIKinCov(Qp_, Qv_, Qw_, pBC_meas-bias_p_, w_meas-bias_imu_.tail(3), Iq);
//
//        // // Compute with the traditionnal method for comparison (unitest to extract later)
//        // Matrix9d J = computeIKinJac(pBC_meas-bias_p_, w_meas-bias_imu_.tail(3), Iq);
//        // Matrix9d QkinFromJ = J * Qn * J.transpose();
//
//        // =============== CREATE CAPURE/FEATURE/FACTOR
//        Vector9d meas; meas << pBC_meas, vBC_meas, w_meas;
//        CaptureBasePtr cap = CaptureBase::emplace<CaptureBase>(KF0_, "CaptureInertialKinematics", t_, nullptr);
//        auto feat = FeatureBase::emplace<FeatureInertialKinematics>(cap, meas, Qkin, Iq, Lq, FeatureBase::UncertaintyType::UNCERTAINTY_IS_COVARIANCE);
//        feat_in_ = static_pointer_cast<FeatureInertialKinematics>(feat); // !! auto feat_in_ creates no compilation error but a segfault
//        // auto fac = FactorBase::emplace<FactorInertialKinematics>(feat_in_, feat_in_, sbbimu_, nullptr, false);
//    }
//
//    virtual void TearDown(){}
//};



////////////////////////////////////////////////////////
// =============== TEST FONCTIONS ======================
////////////////////////////////////////////////////////


TEST_F(FactorInertialKinematics_1KF_Meas0_bpfix,ZeroMvt)
{
    Eigen::Vector3d c_exp  = Eigen::Vector3d::Zero();
    Eigen::Vector3d cd_exp = Eigen::Vector3d::Zero();
    Eigen::Vector3d Lc_exp = Eigen::Vector3d::Zero();
    // Eigen::Vector3d bp_exp = Eigen::Vector3d::Zero();

    KF0_->perturb();
    string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;

    ASSERT_MATRIX_APPROX(KF0_->getP()->getState(), x_origin_.at('P'), 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('C')->getState(), c_exp, 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('D')->getState(), cd_exp, 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('L')->getState(), Lc_exp, 1e-5);
}


TEST_F(FactorInertialKinematics_1KF_1p_bpfix,ZeroMvt)
{
    Eigen::Vector3d c_exp; c_exp << 1,0,0;
    Eigen::Vector3d cd_exp = Eigen::Vector3d::Zero();
    Eigen::Vector3d Lc_exp = Eigen::Vector3d::Zero();
    // Eigen::Vector3d bp_exp = Eigen::Vector3d::Zero();

    KF0_->perturb();
    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;

    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('C')->getState(), c_exp, 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('D')->getState(), cd_exp, 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('L')->getState(), Lc_exp, 1e-5);
}

TEST_F(FactorInertialKinematics_1KF_m1p_pfix,ZeroMvt)
{
    // Eigen::Vector3d c_exp  = Eigen::Vector3d::Zero();
    Eigen::Vector3d cd_exp = Eigen::Vector3d::Zero();
    Eigen::Vector3d Lc_exp = Eigen::Vector3d::Zero();
    Eigen::Vector3d bp_exp; bp_exp << 1,0,0;

    KF0_->perturb();
    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;

    // ASSERT_MATRIX_APPROX(KF0_->getStateBlock('C')->getState(), c_exp, 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('D')->getState(), cd_exp, 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('L')->getState(), Lc_exp, 1e-5);
    ASSERT_MATRIX_APPROX(CIK0_->getStateBlock('I')->getState(), bp_exp, 1e-5);
}

TEST_F(FactorInertialKinematics_1KF_1v_bfix,ZeroMvt)
{
    Eigen::Vector3d c_exp  = Eigen::Vector3d::Zero();
    Eigen::Vector3d cd_exp; cd_exp << 1,0,0;
    Eigen::Vector3d Lc_exp = Eigen::Vector3d::Zero();
    // Eigen::Vector3d bp_exp = Eigen::Vector3d::Zero();

    KF0_->perturb();
    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;

    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('C')->getState(), c_exp, 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('D')->getState(), cd_exp, 1e-5);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('L')->getState(), Lc_exp, 1e-5);
}


int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);

  return RUN_ALL_TESTS();
}
