//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * gtest_force_torque_delta_tools.cpp
 *
 *  Created on: Feb 17, 2020
 *      Author: mfourmy
 */

#include "bodydynamics/math/force_torque_delta_tools.h"
#include <core/utils/utils_gtest.h>

using namespace Eigen;
using namespace wolf;
using namespace bodydynamics;

TEST(FT_tools, identity)
{
    VectorXd id_true;
    id_true.setZero(13);
    id_true(12) = 1.0;

    VectorXd id = identity<double>();
    ASSERT_MATRIX_APPROX(id, id_true, 1e-10);
}

TEST(FT_tools, identity_split)
{
    VectorXd c(3), cd(3), Lc(3), qv(4);
    Quaterniond q;

    identity(c, cd, Lc, q);
    ASSERT_MATRIX_APPROX(c, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(cd, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(Lc, Vector3d::Zero(), 1e-10);
    ASSERT_QUATERNION_APPROX(q, Quaterniond::Identity(), 1e-10);

    identity(c, cd, Lc, qv);
    ASSERT_MATRIX_APPROX(c, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(cd, Vector3d::Zero(), 1e-10);
    ASSERT_MATRIX_APPROX(Lc, Vector3d::Zero(), 1e-10);    
    ASSERT_MATRIX_APPROX(qv, (Vector4d()<<0,0,0,1).finished(), 1e-10);
}

TEST(FT_tools, inverse)
{
    VectorXd d(13), id(13), iid(13), iiid(13), I(13);
    Vector4d qv;
    double dt = 0.1;

    qv = (Vector4d() << 9, 10, 11, 12).finished().normalized();
    d << 0, 1, 2, 3, 4, 5, 6, 7, 8, qv;

    id   = inverse(d, dt);

    compose(id, d, dt, I);
    ASSERT_MATRIX_APPROX(I, identity(), 1e-10);
    compose(d, id, -dt, I); // Observe -dt is reversed !!
    ASSERT_MATRIX_APPROX(I, identity(), 1e-10);

    inverse(id, -dt, iid); // Observe -dt is reversed !!
    ASSERT_MATRIX_APPROX( d,  iid, 1e-10);
    iiid = inverse(iid, dt);
    ASSERT_MATRIX_APPROX(id, iiid, 1e-10);
}

TEST(FT_tools, compose_between)
{
    VectorXd dx1(13), dx2(13), dx3(13);
    Matrix<double, 13, 1> d1, d2, d3;
    double dt = 0.1;

    dx1 << Vector9d::Random(), Vector4d::Random().normalized();
    d1 = dx1;
    dx2 << Vector9d::Random(), Vector4d::Random().normalized();
    d2 = dx2;

    // combinations of templates for sum()
    compose(dx1, dx2, dt, dx3);
    compose(d1, d2, dt, d3);
    ASSERT_MATRIX_APPROX(d3, dx3, 1e-10);

    compose(d1, dx2, dt, dx3);
    d3 = compose(dx1, d2, dt);
    ASSERT_MATRIX_APPROX(d3, dx3, 1e-10);

    // // minus(d1, d3) should recover delta_2
    VectorXd diffx(13);
    Matrix<double,13,1> diff;
    between(d1, d3, dt, diff);
    ASSERT_MATRIX_APPROX(diff, d2, 1e-10);

    // minus(d3, d1, -dt) should recover inverse(d2, dt)
    diff = between(d3, d1, -dt);
    ASSERT_MATRIX_APPROX(diff, inverse(d2, dt), 1e-10);
}

TEST(FT_tools, compose_between_with_state)
{
    VectorXd x(13), d(13), x2(13), x3(13), d2(13), d3(13);
    double dt = 0.1;

    x << Vector9d::Random(), Vector4d::Random().normalized();
    d << Vector9d::Random(), Vector4d::Random().normalized();

    composeOverState(x, d, dt, x2);
    x3 = composeOverState(x, d, dt);
    ASSERT_MATRIX_APPROX(x3, x2, 1e-10);

    // betweenStates(x, x2) should recover d
    betweenStates(x, x2, dt, d2);
    d3 = betweenStates(x, x2, dt);
    ASSERT_MATRIX_APPROX(d2, d, 1e-10);
    ASSERT_MATRIX_APPROX(d3, d, 1e-10);
    ASSERT_MATRIX_APPROX(betweenStates(x, x2, dt), d, 1e-10);

    // x + (x2 - x) = x2
    ASSERT_MATRIX_APPROX(composeOverState(x, betweenStates(x, x2, dt), dt), x2, 1e-10);

    // (x + d) - x = d
    ASSERT_MATRIX_APPROX(betweenStates(x, composeOverState(x, d, dt), dt), d, 1e-10);
}

TEST(FT_tools, lift_retract)
{
    VectorXd d_min(12); d_min << Vector9d::Random(), .1*Vector3d::Random(); // use angles in the ball theta < pi

    // transform to delta
    VectorXd delta = exp_FT(d_min);

    // expected delta
    Vector3d dc = d_min.head(3);
    Vector3d dcd = d_min.segment<3>(3);
    Vector3d dLc = d_min.segment<3>(6);
    Quaterniond dq = v2q(d_min.segment<3>(9));
    VectorXd delta_true(13); delta_true << dc, dcd, dLc, dq.coeffs();
    ASSERT_MATRIX_APPROX(delta, delta_true, 1e-10);

    // transform to a new tangent -- should be the original tangent
    VectorXd d_from_delta = log_FT(delta);
    ASSERT_MATRIX_APPROX(d_from_delta, d_min, 1e-10);

    // transform to a new delta -- should be the previous delta
    VectorXd delta_from_d = exp_FT(d_from_delta);
    ASSERT_MATRIX_APPROX(delta_from_d, delta, 1e-10);
}

TEST(FT_tools, plus)
{
    VectorXd d1(13), d2(13), d3(13);
    VectorXd err(12);
    Vector4d qv = Vector4d::Random().normalized();
    d1 <<  Vector9d::Random(), qv;
    err << Matrix<double, 12, 1>::Random();

    d3.head<3>() = d1.head<3>() + err.head<3>();
    d3.segment<3>(3) = d1.segment<3>(3) + err.segment<3>(3);
    d3.segment<3>(6) = d1.segment<3>(6) + err.segment<3>(6);
    d3.tail<4>() = (Quaterniond(qv.data()) * exp_q(err.tail<3>())).coeffs();

    plus(d1, err, d2);
    ASSERT_MATRIX_APPROX(diff(d3, d2), VectorXd::Zero(12), 1e-10);
}

TEST(FT_tools, diff)
{
    VectorXd d1(13), d2(13);
    VectorXd err(12);
    Vector4d qv = Vector4d::Random().normalized();
    d1 <<  Vector9d::Random(), qv;
    d2 = d1;

    diff(d1, d2, err);
    ASSERT_MATRIX_APPROX(err, VectorXd::Zero(12), 1e-10);
    ASSERT_MATRIX_APPROX(diff(d1, d2), VectorXd::Zero(12), 1e-10);

    VectorXd d3(13);
    d3.setRandom(); d3.tail<4>().normalize();
    err.head<3>() = d3.head<3>() - d1.head<3>();
    err.segment<3>(3) = d3.segment<3>(3) - d1.segment<3>(3);
    err.segment<3>(6) = d3.segment<3>(6) - d1.segment<3>(6);    
    err.tail<3>() = log_q(Quaterniond(d1.data()+9).conjugate()*Quaterniond(d3.data()+9));

    ASSERT_MATRIX_APPROX(err, diff(d1, d3), 1e-10);
}

TEST(FT_tools, compose_jacobians)
{
    VectorXd d1(13), d2(13), d3(13), d1_pert(13), d2_pert(13), d3_pert(13); // deltas
    VectorXd t1(12), t2(12); // tangent elements
    Matrix<double, 12, 12> J1_a, J2_a, J1_n, J2_n;
    Vector4d qv1, qv2;
    double dt = 0.1;
    double dx = 1e-6;
    qv1 = Vector4d::Random().normalized();
    d1 << Vector9d::Random(), qv1;
    qv2 = Vector4d::Random().normalized();
    d2 << Vector9d::Random(), qv2;

    // analytical jacobians
    compose(d1, d2, dt, d3, J1_a, J2_a);

    // numerical jacobians
    for (unsigned int i = 0; i<12; i++)
    {
        t1      . setZero();
        t1(i)   = dx;

        // Jac wrt first delta
        d1_pert = plus(d1, t1);                 //     (d1 + t1)
        d3_pert = compose(d1_pert, d2, dt);     //     (d1 + t1) + d2
        t2      = diff(d3, d3_pert);            //   { (d2 + t1) + d2 } - { d1 + d2 }
        J1_n.col(i) = t2/dx;                    // [ { (d2 + t1) + d2 } - { d1 + d2 } ] / dx

        // Jac wrt second delta
        d2_pert = plus(d2, t1);                 //          (d2 + t1)
        d3_pert = compose(d1, d2_pert, dt);     //     d1 + (d2 + t1)
        t2      = diff(d3, d3_pert);            //   { d1 + (d2 + t1) } - { d1 + d2 }
        J2_n.col(i) = t2/dx;                    // [ { d1 + (d2 + t1) } - { d1 + d2 } ] / dx
    }

    // check that numerical and analytical match
    ASSERT_MATRIX_APPROX(J1_a, J1_n, 1e-4);
    ASSERT_MATRIX_APPROX(J2_a, J2_n, 1e-4);
}

TEST(FT_tools, diff_jacobians)
{
    VectorXd d1(13), d2(13), diff3(12), d1_pert(13), d2_pert(13), diff3_pert(12); // deltas
    VectorXd t1(12), t2(12); // tangents
    Matrix<double, 12, 12> J1_a, J2_a, J1_n, J2_n;
    Vector4d qv1, qv2;
    double dx = 1e-6;
    qv1 = Vector4d::Random().normalized();
    d1 << Vector9d::Random(), qv1;
    qv2 = Vector4d::Random().normalized();
    d2 << Vector9d::Random(), qv2;

    // analytical jacobians
    diff(d1, d2, diff3, J1_a, J2_a);

    // numerical jacobians
    for (unsigned int i = 0; i<12; i++)
    {
        t1      . setZero();
        t1(i)   = dx;

        // Jac wrt first delta
        d1_pert = plus(d1, t1);                 //          (d1 + t1)
        diff3_pert = diff(d1_pert, d2);            //     d2 - (d1 + t1)
        t2      = diff3_pert - diff3;                 //   { d2 - (d1 + t1) } - { d2 - d1 }
        J1_n.col(i) = t2/dx;                    // [ { d2 - (d1 + t1) } - { d2 - d1 } ] / dx

        // Jac wrt second delta
        d2_pert = plus(d2, t1);                 //     (d2 + t1)
        diff3_pert = diff(d1, d2_pert);            //     (d2 + t1) - d1
        t2      = diff3_pert - diff3;                 //   { (d2 + t1) - d1 } - { d2 - d1 }
        J2_n.col(i) = t2/dx;                    // [ { (d2 + t1) - d1 } - { d2 - d1 } ] / dx
    }

    // check that numerical and analytical match
    ASSERT_MATRIX_APPROX(J1_a, J1_n, 1e-4);
    ASSERT_MATRIX_APPROX(J2_a, J2_n, 1e-4);
}

TEST(FT_tools, body2delta_jacobians)
{
    VectorXd delta(13), delta_pert(13); // deltas
    VectorXd body(32), pert(32), body_pert(32);
    VectorXd tang(12); // tangents
    Matrix<double, 12, 30> J_a, J_n; // analytic and numerical jacs
    double dt = 0.1;
    double mass = 1;
    double dx = 1e-6;
    Vector4d qv = Vector4d::Random().normalized();
    delta << Vector9d::Random(), qv;
    Vector4d bql1 = Vector4d::Random().normalized();
    Vector4d bql2 = Vector4d::Random().normalized();
    body << Matrix<double, 24, 1>::Random(), bql1, bql2;  //

    // analytical jacobians
    body2delta(body, dt, mass, 2, 6, delta, J_a);

    // numerical jacobians
    J_n.setZero();
    for (unsigned int i = 0; i<18; i++)
    {
        pert      . setZero();
        pert(i)   = dx;

        // Jac wrt first delta
        body_pert  = body + pert;
        delta_pert = body2delta(body_pert, dt, mass, 2, 6);   //   delta(body + pert)
        tang       = diff(delta, delta_pert);           //   delta(body + pert) -- delta(body)
        J_n.col(i) = tang/dx;                          //  ( delta(body + pert) -- delta(body) ) / dx
    }

    // check that numerical and analytical match
    // ASSERT_MATRIX_APPROX(J_a.block(0,0,12,18), J_n.block(0,0,12,18), 1e-4);  // only compare the implemented parts -> kinematics uncertainty not considered here
    ASSERT_MATRIX_APPROX(J_a, J_n, 1e-4);
}


TEST(FT_tools, debiasData_jacobians)
{
    VectorXd data(32), bias(6), bias_pert(6), pert(6), body(32), body_pert(32), tang(32);
    Matrix<double, 30, 6> J_a, J_n; // analytic and numerical jacs
    double dx = 1e-6;
    Vector4d bql1 = Vector4d::Random().normalized();
    Vector4d bql2 = Vector4d::Random().normalized();
    data << Matrix<double, 24, 1>::Random(), bql1, bql2;  //
    bias << Vector6d::Random();

    // analytical jacobians
    debiasData(data, bias, 2, 6, body, J_a);

    // numerical jacobian
    J_n.setZero();
    for (unsigned int i = 0; i<6; i++)
    {
        pert      . setZero();
        pert(i)   = dx;

        // Jac wrt first delta
        bias_pert  = bias + pert;
        debiasData(data, bias_pert, 2, 6, body_pert);
        // The next line is actually a cheating: 
        // we know that the quaternions at the end of the body vector are not influenced by the 
        // bias perturbation so we just take the euclidean difference of the 2 vectors and remove the last 2 elements
        // to get the tangent space vector
        tang       = (body_pert - body).segment<30>(0);

        J_n.col(i) = tang/dx;
    }

    // check that numerical and analytical match
    ASSERT_MATRIX_APPROX(J_a, J_n, 1e-4);
}



// TEST(motion2data, zero)
// {
//     Vector6d motion;
//     Vector6d bias;
//     Quaterniond q;

//     motion  .setZero();
//     bias    .setZero();
//     q       .setIdentity();

//     Vector6d data = imu::motion2data(motion, q, bias);

//     Vector6d data_true; data_true << -gravity(), Vector3d::Zero();

//     ASSERT_MATRIX_APPROX(data, data_true, 1e-12);
// }

// TEST(motion2data, motion)
// {
//     Vector6d motion, g_extended;
//     Vector6d bias;
//     Quaterniond q;

//     g_extended << gravity() , Vector3d::Zero();

//     motion  << 1,2,3, 4,5,6;
//     bias    .setZero();
//     q       .setIdentity();

//     Vector6d data = imu::motion2data(motion, q, bias);

//     Vector6d data_true; data_true = motion - g_extended;

//     ASSERT_MATRIX_APPROX(data, data_true, 1e-12);
// }

// TEST(motion2data, bias)
// {
//     Vector6d motion, g_extended;
//     Vector6d bias;
//     Quaterniond q;

//     g_extended << gravity() , Vector3d::Zero();

//     motion  .setZero();
//     bias    << 1,2,3, 4,5,6;
//     q       .setIdentity();

//     Vector6d data = imu::motion2data(motion, q, bias);

//     Vector6d data_true; data_true = bias - g_extended;

//     ASSERT_MATRIX_APPROX(data, data_true, 1e-12);
// }

// TEST(motion2data, orientation)
// {
//     Vector6d motion, g_extended;
//     Vector6d bias;
//     Quaterniond q;

//     g_extended << gravity() , Vector3d::Zero();

//     motion  .setZero();
//     bias    .setZero();
//     q       = v2q(Vector3d(M_PI/2, 0, 0)); // turn 90 deg in X axis

//     Vector6d data = imu::motion2data(motion, q, bias);

//     Vector6d data_true; data_true << 0,-gravity()(2),0, 0,0,0;

//     ASSERT_MATRIX_APPROX(data, data_true, 1e-12);
// }

// TEST(motion2data, AllRandom)
// {
//     Vector6d motion, g_extended;
//     Vector6d bias;
//     Quaterniond q;

//     motion      .setRandom();
//     bias        .setRandom();
//     q.coeffs()  .setRandom().normalize();

//     g_extended << q.conjugate()*gravity() , Vector3d::Zero();

//     Vector6d data = imu::motion2data(motion, q, bias);

//     Vector6d data_true; data_true = motion + bias - g_extended;

//     ASSERT_MATRIX_APPROX(data, data_true, 1e-12);
// }

// /* Integrate acc and angVel motion, obtain Delta_preintegrated
//  * Input:
//  *   N: number of steps
//  *   q0: initial orientaiton
//  *   motion: [ax, ay, az, wx, wy, wz] as the true magnitudes in body brame
//  *   bias_real: the real bias of the Imu
//  *   bias_preint: the bias used for Delta pre-integration
//  * Output:
//  *   return: the preintegrated delta
//  */
// VectorXd integrateDelta(int N, const Quaterniond& q0, const VectorXd& motion, const VectorXd& bias_real, const VectorXd& bias_preint, double dt)
// {
//     VectorXd data(6);
//     VectorXd body(6);
//     VectorXd delta(13);
//     VectorXd Delta(13), Delta_plus(13);
//     Delta << 0,0,0, 0,0,0,1, 0,0,0;
//     Quaterniond q(q0);
//     for (int n = 0; n<N; n++)
//     {
//         data        = motion2data(motion, q, bias_real);
//         q           = q*exp_q(motion.tail(3)*dt);
//         body        = data - bias_preint;
//         delta       = body2delta(body, dt);
//         Delta_plus  = compose(Delta, delta, dt);
//         Delta       = Delta_plus;
//     }
//     return Delta;
// }

// /* Integrate acc and angVel motion, obtain Delta_preintegrated
//  * Input:
//  *   N: number of steps
//  *   q0: initial orientaiton
//  *   motion: [ax, ay, az, wx, wy, wz] as the true magnitudes in body brame
//  *   bias_real: the real bias of the Imu
//  *   bias_preint: the bias used for Delta pre-integration
//  * Output:
//  *   J_D_b: the Jacobian of the preintegrated delta wrt the bias
//  *   return: the preintegrated delta
//  */
// VectorXd integrateDelta(int N, const Quaterniond& q0, const VectorXd& motion, const VectorXd& bias_real, const VectorXd& bias_preint, double dt, Matrix<double, 9, 6>& J_D_b)
// {
//     VectorXd data(6);
//     VectorXd body(6);
//     VectorXd delta(13);
//     Matrix<double, 9, 6> J_d_d, J_d_b;
//     Matrix<double, 9, 9> J_D_D, J_D_d;
//     VectorXd Delta(13), Delta_plus(13);
//     Quaterniond q;

//     Delta << 0,0,0, 0,0,0,1, 0,0,0;
//     J_D_b.setZero();
//     q = q0;
//     for (int n = 0; n<N; n++)
//     {
//         // Simulate data
//         data = motion2data(motion, q, bias_real);
//         q    = q*exp_q(motion.tail(3)*dt);
//         // Motion::integrateOneStep()
//         {   // Imu::computeCurrentDelta
//             body  = data - bias_preint;
//             body2delta(body, dt, delta, J_d_d);
//             J_d_b = - J_d_d;
//         }
//         {   // Imu::deltaPlusDelta
//             compose(Delta, delta, dt, Delta_plus, J_D_D, J_D_d);
//         }
//         // Motion:: jac calib
//         J_D_b = J_D_D*J_D_b + J_D_d*J_d_b;
//         // Motion:: buffer
//         Delta = Delta_plus;
//     }
//     return Delta;
// }

// TEST(FT_tools, integral_jacobian_bias)
// {
//     VectorXd Delta(13), Delta_pert(13); // deltas
//     VectorXd bias_real(6), pert(6), bias_pert(6), bias_preint(6);
//     VectorXd body(6), data(6), motion(6);
//     VectorXd tang(9); // tangents
//     Matrix<double, 9, 6> J_a, J_n; // analytic and numerical jacs
//     double dt = 0.001;
//     double dx = 1e-4;
//     Quaterniond q0(3, 4, 5, 6); q0.normalize();
//     motion << .1, .2, .3,   .3, .2, .1;
//     bias_real   << .002, .004, .001,   .003, .002, .001;
//     bias_preint << .004, .005, .006,   .001, .002, .003;

//     int N = 500; // # steps of integration

//     // Analytical Jacobians
//     Delta = integrateDelta(N, q0, motion, bias_real, bias_preint, dt, J_a);

//     // numerical jacobians
//     for (unsigned int i = 0; i<6; i++)
//     {
//         pert       . setZero();
//         pert(i)    = dx;

//         bias_pert  = bias_preint + pert;
//         Delta_pert = integrateDelta(N, q0, motion, bias_real, bias_pert, dt);
//         tang       = diff(Delta, Delta_pert);       //   Delta(bias + pert) -- Delta(bias)
//         J_n.col(i) = tang/dx;                       // ( Delta(bias + pert) -- Delta(bias) ) / dx
//     }

//     // check that numerical and analytical match
//     ASSERT_MATRIX_APPROX(J_a, J_n, 1e-4);
// }

// TEST(FT_tools, delta_correction)
// {
//     VectorXd Delta(13), Delta_preint(13), Delta_corr; // deltas
//     VectorXd bias(6), pert(6), bias_preint(6);
//     VectorXd body(6), motion(6);
//     VectorXd tang(9); // tangents
//     Matrix<double, 9, 6> J_b; // analytic and numerical jacs
//     Vector4d qv;;
//     double dt = 0.001;
//     Quaterniond q0(3, 4, 5, 6); q0.normalize();
//     motion << 1, 2, 3,   3, 2, 1; motion *= .1;

//     int N = 1000; // # steps of integration

//     // preintegration with correct bias
//     bias << .004, .005, .006,   .001, .002, .003;
//     Delta = integrateDelta(N, q0, motion, bias, bias, dt);

//     // preintegration with wrong bias
//     pert << .002, -.001, .003,   -.0003, .0002, -.0001;
//     bias_preint = bias + pert;
//     Delta_preint = integrateDelta(N, q0, motion, bias, bias_preint, dt, J_b);

//     // correct perturbated
//     Vector9d step = J_b*(bias-bias_preint);
//     Delta_corr = plus(Delta_preint, step);

//     // Corrected delta should match real delta
//     ASSERT_MATRIX_APPROX(Delta, Delta_corr, 1e-5);

//     // diff between real and corrected should be zero
//     ASSERT_MATRIX_APPROX(diff(Delta, Delta_corr), Vector9d::Zero(), 1e-5);

//     // diff between preint and corrected deltas should be the jacobian-computed step
//     ASSERT_MATRIX_APPROX(diff(Delta_preint, Delta_corr), step, 1e-5);
// }

// TEST(FT_tools, full_delta_residual)
// {
//     VectorXd x1(13), x2(13);
//     VectorXd Delta(13), Delta_preint(13), Delta_corr(13);
//     VectorXd Delta_real(9), Delta_err(9), Delta_diff(13), Delta_exp(13);
//     VectorXd bias(6), pert(6), bias_preint(6), bias_null(6);
//     VectorXd body(6), motion(6);
//     VectorXd tang(9); // tangents
//     Matrix<double, 9, 6> J_b; // analytic and numerical jacs
//     double dt = 0.001;
//     Quaterniond q0; q0.setIdentity();

//     x1          << 0,0,0, 0,0,0,1, 0,0,0;
//     motion      <<  .3,    .2,    .1,      .1,     .2,     .3; // acc and gyro
// //    motion      <<  .3,    .2,    .1,      .0,     .0,     .0; // only acc
// //    motion      <<  .0,    .0,    .0,      .1,     .2,     .3; // only gyro
//     bias        << 0.01, 0.02, 0.003,   0.002, 0.005, 0.01;
//     bias_null   << 0,     0,     0,       0,      0,      0;
// //    bias_preint << 0.003, 0.002, 0.001,   0.001,  0.002,  0.003;
//     bias_preint = bias_null;

//     int N = 1000; // # steps of integration

//     // preintegration with no bias
//     Delta_real = integrateDelta(N, q0, motion, bias_null, bias_null, dt);

//     // final state
//     x2 = composeOverState(x1, Delta_real, 1000*dt);

//     // preintegration with the correct bias
//     Delta = integrateDelta(N, q0, motion, bias, bias, dt);

//     ASSERT_MATRIX_APPROX(Delta, Delta_real, 1e-4);

//     // preintegration with wrong bias
//     Delta_preint = integrateDelta(N, q0, motion, bias, bias_preint, dt, J_b);

//     // compute correction step
//     Vector9d step = J_b*(bias-bias_preint);

//     // correct preintegrated delta
//     Delta_corr = plus(Delta_preint, step);

//     // Corrected delta should match real delta
//     ASSERT_MATRIX_APPROX(Delta_real, Delta_corr, 1e-3);

//     // diff between real and corrected should be zero
//     ASSERT_MATRIX_APPROX(diff(Delta_real, Delta_corr), Vector9d::Zero(), 1e-3);

//     // expected delta
//     Delta_exp = betweenStates(x1, x2, N*dt);

//     ASSERT_MATRIX_APPROX(Delta_exp, Delta_corr, 1e-3);

//     // compute diff between expected and corrected
// //    Matrix<double, 9, 9> J_err_exp, J_err_corr;
//     diff(Delta_exp, Delta_corr, Delta_err); //, J_err_exp, J_err_corr);

//     ASSERT_MATRIX_APPROX(Delta_err, Vector9d::Zero(), 1e-3);

//     // compute error between expected and preintegrated
//     Delta_err = diff(Delta_preint, Delta_exp);

//     // diff between preint and corrected deltas should be the jacobian-computed step
//     ASSERT_MATRIX_APPROX(diff(Delta_preint, Delta_corr), step, 1e-3);
//     /* This implies:
//      *   - Since D_corr is expected to be similar to D_exp
//      *      step ~ diff(Delta_preint, Delta_exp)
//      *   - the residual can be computed with a regular '-' :
//      *      residual = diff(D_preint, D_exp) - J_bias * (bias - bias_preint)
//      */

// //    WOLF_TRACE("Delta real: ", Delta_real.transpose());
// //    WOLF_TRACE("x2: ", x2.transpose());
// //    WOLF_TRACE("Delta: ", Delta.transpose());
// //    WOLF_TRACE("Delta pre: ", Delta_preint.transpose());
// //    WOLF_TRACE("Jac bias: \n", J_b);
// //    WOLF_TRACE("Delta step: ", step.transpose());
// //    WOLF_TRACE("Delta cor: ", Delta_corr.transpose());
// //    WOLF_TRACE("Delta exp: ", Delta_exp.transpose());
// //    WOLF_TRACE("Delta err: ", Delta_err.transpose());
// //    WOLF_TRACE("Delta err exp-pre: ", Delta_err.transpose());
// }

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
//  ::testing::GTEST_FLAG(filter) = "FT_tools.delta_correction";
  return RUN_ALL_TESTS();
}

