//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

/**
 * \file gtest_factor_force_torque.cpp
 *
 *  Created on: March 20, 2020
 *      \author: Mederic Fourmy
 */

// debug
#include <iostream>
#include <fstream>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>

// WOLF
#include <core/problem/problem.h>
#include <core/ceres_wrapper/solver_ceres.h>

// #include <core/ceres_wrapper/qr_manager.h>

#include <core/math/rotations.h>
#include <core/capture/capture_base.h>
#include <core/capture/capture_pose.h>
#include <core/feature/feature_base.h>
#include <core/factor/factor_block_absolute.h>
#include <core/factor/factor_relative_pose_3d.h>

// IMU
#include "imu/sensor/sensor_imu.h"
#include "imu/capture/capture_imu.h"
#include "imu/processor/processor_imu.h"

// BODYDYNAMICS
#include "bodydynamics/internal/config.h"
#include "bodydynamics/sensor/sensor_inertial_kinematics.h"
#include "bodydynamics/sensor/sensor_force_torque.h"
#include "bodydynamics/capture/capture_inertial_kinematics.h"
#include "bodydynamics/capture/capture_force_torque.h"
#include "bodydynamics/processor/processor_inertial_kinematics.h"
#include "bodydynamics/processor/processor_force_torque.h"

#include <Eigen/Eigenvalues>

using namespace wolf;
using namespace Eigen;
using namespace std;

const Vector3d ZERO3 = Vector3d::Zero();
const Vector6d ZERO6 = Vector6d::Zero();
const Vector3d ONES3 = Vector3d::Ones();

const Vector3d BIAS_PBC_SMAL = {0.01, 0.02, 0.03};

const double ACC = 1;
const double MASS = 1;
const double DT = 1.0;
const Vector3d PBCX = {-0.1, 0, 0};  // X axis offset 
const Vector3d PBCY = {0, -0.1, 0};  // Y axis offset 
const Vector3d PBCZ = {0, 0, -0.1};  // Z axis offset


class ProcessorForceTorqueImuOdom3dIkin2KF : public testing::Test
{
public:
    wolf::TimeStamp t0_;
    ProblemPtr problem_;
    SensorImuPtr sen_imu_;
    SolverCeresPtr solver_;
    SensorInertialKinematicsPtr sen_ikin_;
    SensorForceTorquePtr sen_ft_;
    ProcessorImuPtr proc_imu_;
    ProcessorInertialKinematicsPtr proc_inkin_;
    ProcessorForceTorquePtr proc_ft_;
    FrameBasePtr KF1_;

    void SetUp() override
    {
        std::string bodydynamics_root_dir = _WOLF_BODYDYNAMICS_ROOT_DIR;

        problem_ = Problem::create("POVCDL", 3);

        // CERES WRAPPER
        solver_ = std::make_shared<SolverCeres>(problem_);
        // solver_->getSolverOptions().minimizer_type = ceres::TRUST_REGION; //ceres::TRUST_REGION;ceres::LINE_SEARCH
        // solver_->getSolverOptions().max_line_search_step_contraction = 1e-3;
        // solver_->getSolverOptions().max_num_iterations = 1e4;

        //===================================================== INITIALIZATION
        // SENSOR + PROCESSOR INERTIAL KINEMATICS
        ParamsSensorInertialKinematicsPtr intr_ik = std::make_shared<ParamsSensorInertialKinematics>();
        intr_ik->std_pbc = 0.1;
        intr_ik->std_vbc = 0.1;
        VectorXd extr; // default size for dynamic vector is 0-0 -> what we need
        SensorBasePtr sen_ikin_base = problem_->installSensor("SensorInertialKinematics", "SenIK", extr, intr_ik);
        // SensorBasePtr sen_ikin_base = problem_->installSensor("SensorInertialKinematics", "SenIK", extr,  bodydynamics_root_dir + "/demos/sensor_inertial_kinematics.yaml");  // TODO: does not work!
        sen_ikin_ = std::static_pointer_cast<SensorInertialKinematics>(sen_ikin_base);

        ParamsProcessorInertialKinematicsPtr params_ik = std::make_shared<ParamsProcessorInertialKinematics>();
        params_ik->sensor_angvel_name = "SenImu";
        params_ik->std_bp_drift = 1e-2;
        ProcessorBasePtr proc_ikin_base = problem_->installProcessor("ProcessorInertialKinematics", "PInertialKinematics", sen_ikin_, params_ik);
        // ProcessorBasePtr proc_ikin_base = problem_->installProcessor("ProcessorInertialKinematics", "PInertialKinematics", "SenIK", bodydynamics_root_dir + "/test/processor_inertial_kinematics.yaml");
        proc_inkin_ = std::static_pointer_cast<ProcessorInertialKinematics>(proc_ikin_base);

        // SENSOR + PROCESSOR IMU
        SensorBasePtr sen_imu_base = problem_->installSensor("SensorImu", "SenImu", (Vector7d() << 0, 0, 0, 0, 0, 0, 1).finished(), bodydynamics_root_dir + "/test/sensor_imu.yaml");
        sen_imu_ = std::static_pointer_cast<SensorImu>(sen_imu_base);
        ProcessorBasePtr proc_ftp_base = problem_->installProcessor("ProcessorImu", "imu pre-integrator", "SenImu", bodydynamics_root_dir + "/test/processor_imu.yaml");
        proc_imu_ = std::static_pointer_cast<ProcessorImu>(proc_ftp_base);

        // SENSOR + PROCESSOR FT PREINT
        ParamsSensorForceTorquePtr intr_ft = std::make_shared<ParamsSensorForceTorque>();
        intr_ft->std_f = 0.001;
        intr_ft->std_tau = 0.001;
        intr_ft->mass = MASS;
        SensorBasePtr sen_ft_base = problem_->installSensor("SensorForceTorque", "SenFT", VectorXd(0), intr_ft);
        // SensorBasePtr sen_ft_base = problem_->installSensor("SensorForceTorque", "SenFT", VectorXd(0), bodydynamics_root_dir + "/demos/sensor_ft.yaml");
        sen_ft_ = std::static_pointer_cast<SensorForceTorque>(sen_ft_base);
        ParamsProcessorForceTorquePtr params_sen_ft = std::make_shared<ParamsProcessorForceTorque>();
        params_sen_ft->sensor_ikin_name = "SenIK";
        params_sen_ft->sensor_angvel_name = "SenImu";
        params_sen_ft->nbc = 2;
        params_sen_ft->dimc = 6;
        params_sen_ft->time_tolerance = 0.0005;
        params_sen_ft->unmeasured_perturbation_std = 1e-4;
        params_sen_ft->max_time_span = 1000;
        params_sen_ft->max_buff_length = 500;
        params_sen_ft->dist_traveled = 20000.0;
        params_sen_ft->angle_turned = 1000;
        params_sen_ft->voting_active = true;
        ProcessorBasePtr proc_ft_base = problem_->installProcessor("ProcessorForceTorque", "PFT", sen_ft_, params_sen_ft);
        // ProcessorBasePtr proc_ft_base = problem_->installProcessor("ProcessorForceTorque", "PFT", "SenFT", bodydynamics_root_dir + "/demos/processor_ft_preint");
        proc_ft_ = std::static_pointer_cast<ProcessorForceTorque>(proc_ft_base);
    }

    void TearDown() override {}
};


class Cst2KFZeroMotion : public ProcessorForceTorqueImuOdom3dIkin2KF
{
public:
    FrameBasePtr KF2_;

    // Unitary data
    Vector3d acc_meas_;
    Vector3d w_meas_;
    Vector3d pbc_;
    Vector3d pbc_meas_;
    Vector3d vBC_meas_;
    Vector3d f1_;
    Vector3d tau1_;
    Vector3d pbl1_;
    Vector4d bql1_;
    Vector3d f2_;
    Vector3d tau2_;
    Vector3d pbl2_;
    Vector4d bql2_;

    Matrix6d cov_f_;
    Matrix6d cov_tau_;
    Matrix3d cov_pbc_;
    Matrix3d cov_wb_;
    
    // Aggregated data to construct Captures 
    Vector6d acc_gyr_meas_;
    Matrix6d acc_gyr_cov_;
    Vector9d meas_ikin_;
    MatrixXd cap_ftp_data_;
    MatrixXd Qftp_;
    Matrix3d Iq_;
    Vector3d Lq_;
    Vector3d bias_pbc_;
    Vector7d prev_pose_curr_;

    VectorXd x_origin_;



    Cst2KFZeroMotion()
    {
        bias_pbc_ = ZERO3;
    }

    virtual void setOriginState()
    {
        x_origin_.resize(19);
        x_origin_ << ZERO3, 0, 0, 0, 1, ZERO3, ZERO3, ZERO3, ZERO3;
    }

    void SetUp() override
    {
        ProcessorForceTorqueImuOdom3dIkin2KF::SetUp();
        t0_.set(0.0);
        TimeStamp t1; t1.set(1*DT);
        TimeStamp t2; t2.set(2*DT);
        TimeStamp t3; t3.set(3*DT);
        TimeStamp t4; t4.set(3*DT);

        // SETPRIOR RETRO-ENGINEERING
        // We are not using setPrior because of processors initial captures problems so we have to
        // - Manually set problem prior
        // - call setOrigin on processors MotionProvider
        setOriginState();
        MatrixXd P_origin_ = pow(1e-3, 2) * MatrixXd::Identity(18,18);
        KF1_ = problem_->emplaceFrame( t0_, x_origin_);
        // Prior pose factor
        CapturePosePtr pose_prior_capture = CaptureBase::emplace<CapturePose>(KF1_, t0_, nullptr, x_origin_.head(7), P_origin_.topLeftCorner(6, 6));
        pose_prior_capture->emplaceFeatureAndFactor();
        ///////////////////////////////////////////////////
        // Prior velocity factor
        CaptureBasePtr capV0 = CaptureBase::emplace<CaptureBase>(KF1_, "Vel0", t0_);
        FeatureBasePtr featV0 = FeatureBase::emplace<FeatureBase>(capV0, "Vel0", x_origin_.segment<3>(7), P_origin_.block<3, 3>(6, 6));
        FactorBasePtr facV0 = FactorBase::emplace<FactorBlockAbsolute>(featV0, featV0, KF1_->getV(), nullptr, false);

        initializeData1();

        // Special trick to make things work in the IMU + IKin + FTPreint case
        // 0. Call keyFrameCallback of processorIKin so that its kf_buffer contains the first KF0
        // 1. Call setOrigin processorIMU  to generate a fake captureIMU -> generates b_imu
        // 2. process one IKin capture corresponding to the same KF, the factor needing b_imu -> generates a b_pbc
        // 3. set processorForceTorquePreint origin which requires both b_imu and b_pbc
        proc_inkin_->keyFrameCallback(KF1_, 0.5);
        proc_imu_->setOrigin(KF1_);
        auto CIKin0 = std::make_shared<CaptureInertialKinematics>(t0_, sen_ikin_, meas_ikin_, Iq_, Lq_);
        CIKin0->process();
        proc_ft_->setOrigin(KF1_);

        // T1
        CaptureImuPtr CImu1 = std::make_shared<CaptureImu>(t1, sen_imu_, acc_gyr_meas_, acc_gyr_cov_);
        CImu1->process();
        auto CIKin1 = std::make_shared<CaptureInertialKinematics>(t1, sen_ikin_, meas_ikin_, Iq_, Lq_);
        CIKin1->process();
        auto CFTpreint1 = std::make_shared<CaptureForceTorque>(t1, sen_ft_, CIKin1, CImu1, cap_ftp_data_, Qftp_);
        CFTpreint1->process();


        changeForData2();


        // T2
        CaptureImuPtr CImu2 = std::make_shared<CaptureImu>(t2, sen_imu_, acc_gyr_meas_, acc_gyr_cov_);
        CImu2->process();
        auto CIKin2 = std::make_shared<CaptureInertialKinematics>(t2, sen_ikin_, meas_ikin_, Iq_, Lq_);
        CIKin2->process();
        auto CFTpreint2 = std::make_shared<CaptureForceTorque>(t2, sen_ft_, CIKin2, CImu2, cap_ftp_data_, Qftp_);
        CFTpreint2->process();

        changeForData3();

        // T3
        CaptureImuPtr CImu3 = std::make_shared<CaptureImu>(t3, sen_imu_, acc_gyr_meas_, acc_gyr_cov_);
        CImu3->process();
        CaptureInertialKinematicsPtr CIKin3 = std::make_shared<CaptureInertialKinematics>(t3, sen_ikin_, meas_ikin_, Iq_, Lq_);
        CIKin3->process();
        auto CFTpreint3 = std::make_shared<CaptureForceTorque>(t3, sen_ft_, CIKin3, CImu3, cap_ftp_data_, Qftp_);
        CFTpreint3->process();

        // T4, just here to make sure the KF at t3 is created
        CaptureImuPtr CImu4 = std::make_shared<CaptureImu>(t4, sen_imu_, acc_gyr_meas_, acc_gyr_cov_);
        CImu4->process();
        CaptureInertialKinematicsPtr CIKin4 = std::make_shared<CaptureInertialKinematics>(t4, sen_ikin_, meas_ikin_, Iq_, Lq_);
        CIKin4->process();
        auto CFTpreint4 = std::make_shared<CaptureForceTorque>(t4, sen_ft_, CIKin3, CImu3, cap_ftp_data_, Qftp_);
        CFTpreint4->process();

        /////////////////////////////////////////////
        // Add an Odom3D to make the biases observable
        setOdomData();
        Matrix6d rel_pose_cov = 1e-6 * Matrix6d::Identity();

        KF2_ = problem_->getTrajectory()->getLastFrame();
        CaptureBasePtr cap_pose_base = CaptureBase::emplace<CapturePose>(KF2_, KF2_->getTimeStamp(), nullptr, prev_pose_curr_, rel_pose_cov);
        FeatureBasePtr ftr_odom3d_base = FeatureBase::emplace<FeaturePose>(cap_pose_base, prev_pose_curr_, rel_pose_cov);
        FactorBase::emplace<FactorRelativePose3d>(ftr_odom3d_base, ftr_odom3d_base, KF1_, nullptr, false, TOP_MOTION);

        ////////////////////////////////////////////
        // Add absolute factor on Imu biases to simulate a fix()
        Matrix6d Q_bi = 1e-8 * Matrix6d::Identity();
        CaptureBasePtr capbi0 = CaptureBase::emplace<CaptureBase>(KF1_, "bi0", t0_);
        FeatureBasePtr featbi0 = FeatureBase::emplace<FeatureBase>(capbi0, "bi0", ZERO6, Q_bi);
        FactorBasePtr facbi0 = FactorBase::emplace<FactorBlockAbsolute>(featbi0, featbi0, KF1_->getCaptureOf(sen_imu_)->getSensorIntrinsic(),nullptr,false);

        // Add loose absolute factor on initial bp bias since dynamical trajectories 
        // able to excite the system enough to make COM and bias_p observable hard to generate in a simple gtest
        Matrix3d Q_bp = 1e-1 * Matrix3d::Identity();
        CaptureBasePtr cap_bp0 = CaptureBase::emplace<CaptureBase>(KF1_, "bp0", t0_);
        FeatureBasePtr feat_bp0 = FeatureBase::emplace<FeatureBase>(cap_bp0, "bp0", bias_pbc_, Q_bp);
        // FeatureBasePtr feat_bp0 = FeatureBase::emplace<FeatureBase>(cap_bp0, "bp0", bias_pbc_ + (Vector3d()<<0.1,0.1,0.1).finished(), Q_bp);
        FactorBasePtr fac_bp0 = FactorBase::emplace<FactorBlockAbsolute>(feat_bp0, feat_bp0, KF1_->getCaptureOf(sen_ikin_)->getSensorIntrinsic(),nullptr,false);
 
    }

    void buildDataVectors(){
        pbc_meas_ = pbc_ + bias_pbc_;

        acc_gyr_meas_ << acc_meas_, w_meas_;
        acc_gyr_cov_ = 1e-3 * Matrix6d::Identity();
        meas_ikin_ << pbc_meas_, vBC_meas_, w_meas_;

        Vector6d f_meas; f_meas << f1_, f2_;
        Vector6d tau_meas; tau_meas << tau1_, tau2_;
        Matrix<double, 14, 1> kin_meas;
        kin_meas << pbl1_, pbl2_, bql1_, bql2_;
        cap_ftp_data_.resize(32,1);
        cap_ftp_data_ << f_meas, tau_meas, pbc_meas_, acc_gyr_meas_.tail<3>(), kin_meas;

        Qftp_ = Matrix<double, 18, 18>::Identity();
        Qftp_.block<6, 6>(0, 0) = cov_f_;
        Qftp_.block<6, 6>(6, 6) = cov_tau_;
        Qftp_.block<3, 3>(12, 12) = cov_pbc_;
        Qftp_.block<3, 3>(15, 15) = cov_wb_;
    }

    void setZeroMotionData()
    {
        // IMU CAPTURE DATA ZERO MOTION
        acc_meas_ = -gravity();
        w_meas_ = ZERO3;

        // INERTIAL KINEMATICS CAPTURE DATA ZERO MOTION
        pbc_ = ZERO3;
        vBC_meas_ = ZERO3;
        // momentum parameters
        Iq_.setIdentity();
        Lq_.setZero();

        // FORCE TORQUE CAPTURE DATA ZERO MOTION
        f1_ << -MASS * wolf::gravity() / 2; // Only gravity -> static robot on both feet
        tau1_ << ZERO3;
        pbl1_ << ZERO3;
        bql1_ << ZERO3, 1;
        f2_ << -MASS * wolf::gravity() / 2; // Only gravity -> static robot on both feet
        tau2_ << ZERO3;
        pbl2_ << ZERO3;
        bql2_ << ZERO3, 1;

        cov_f_ = 1e-4 * Matrix6d::Identity();
        cov_tau_ = 1e-4 * Matrix6d::Identity();
        cov_pbc_ = 1e-4 * Matrix3d::Identity();
        cov_wb_ = 1e-4 * Matrix3d::Identity();
    }

    virtual void initializeData1()
    {
        setZeroMotionData();
        buildDataVectors();
    }

    virtual void changeForData2()
    {
        // DO NOTHING IN BASE CLASS
    }

    virtual void changeForData3()
    {
        // DO NOTHING IN BASE CLASS
    }

    virtual void setOdomData()
    {
        prev_pose_curr_ << ZERO6, 1;
    }

};

class Cst2KFZeroMotionBias : public Cst2KFZeroMotion
{
public:
    void SetUp() override
    {
        bias_pbc_ = BIAS_PBC_SMAL;
        Cst2KFZeroMotion::SetUp();
    }
};

class Cst2KFXaxisLinearMotion : public Cst2KFZeroMotion
{
public:
    void initializeData1() override
    {
        setZeroMotionData();
        acc_meas_[0] += ACC;
        f1_[0] += sen_ft_->getMass() * ACC / 2;
        f2_[0] += sen_ft_->getMass() * ACC / 2;
        // taus according to static Euler eq are zero since no lever

        buildDataVectors();
    }

    void setOdomData() override
    {
        prev_pose_curr_ << 0.5*ACC*pow(3*DT,2), 0, 0, 0, 0, 0, 1;
    }
};

class Cst2KFXaxisLinearMotionPbc : public Cst2KFZeroMotion
{
public:

    void SetUp() override
    {
        Cst2KFZeroMotion::SetUp();
    }

    void setOriginState() override
    {
        Cst2KFZeroMotion::setOriginState();
        x_origin_.segment<3>(10) = PBCX + PBCY + PBCZ; // Ok if PO is 0,0,0, 0,0,0,1
    }

    void initializeData1() override
    {
        setZeroMotionData();
        pbc_ = PBCX + PBCY + PBCZ;
        acc_meas_[0] += ACC;
        f1_[0] += sen_ft_->getMass() * ACC / 2;
        f2_[0] += sen_ft_->getMass() * ACC / 2;
        // taus according to static Euler eq
        tau1_ = -(pbl1_- pbc_).cross(f1_);  // torque at C due to f1
        tau2_ = -(pbl2_- pbc_).cross(f2_);  // torque at C due to f2

        buildDataVectors();
    }

    void setOdomData() override
    {
        prev_pose_curr_ << 0.5*ACC*pow(3*DT,2), 0, 0, 0, 0, 0, 1;
    }
};

class Cst2KFXaxisLinearMotionPbcBias : public Cst2KFZeroMotion
{
public:
    void SetUp() override
    {
        bias_pbc_ = BIAS_PBC_SMAL;
        Cst2KFZeroMotion::SetUp();
    }

    void setOriginState() override
    {
        Cst2KFZeroMotion::setOriginState();
        x_origin_.segment<3>(10) = PBCX + PBCY + PBCZ; // Ok if PO is 0,0,0, 0,0,0,1
    }

    void initializeData1() override
    {
        setZeroMotionData();
        pbc_ = PBCX + PBCY + PBCZ;
        acc_meas_[0] += ACC;
        f1_[0] += sen_ft_->getMass() * ACC / 2;
        f2_[0] += sen_ft_->getMass() * ACC / 2;
        // taus according to static Euler eq
        tau1_ = -(pbl1_- pbc_).cross(f1_);  // torque at C due to f1
        tau2_ = -(pbl2_- pbc_).cross(f2_);  // torque at C due to f2

        buildDataVectors();
    }

    void setOdomData() override
    {
        prev_pose_curr_ << 0.5*ACC*pow(3*DT,2), 0, 0, 0, 0, 0, 1;
    }
};
class Cst2KFXaxisLinearMotionPbcBiasPQbl : public Cst2KFZeroMotion
{
public:
    void SetUp() override
    {
        bias_pbc_ = BIAS_PBC_SMAL;
        Cst2KFZeroMotion::SetUp();
    }

    void setOriginState() override
    {
        Cst2KFZeroMotion::setOriginState();
        x_origin_.segment<3>(10) = PBCX + PBCY + PBCZ; // Ok if PO is 0,0,0, 0,0,0,1
    }

    void initializeData1() override
    {
        setZeroMotionData();
        pbc_ = PBCX + PBCY + PBCZ;
        acc_meas_[0] += ACC;
        f1_[0] += sen_ft_->getMass() * ACC / 2;
        f2_[0] += sen_ft_->getMass() * ACC / 2;
        pbl1_ << 0.1, 0.1, 0.1;
        pbl2_ << 0.1, 0.1, 0.1;
        bql1_ << 1, 0, 0, 0;
        bql2_ << 1, 0, 0, 0;
        Quaterniond quat_bl1(bql1_);
        Quaterniond quat_bl2(bql2_);
        f1_ = quat_bl1.inverse() * f1_;
        f2_ = quat_bl2.inverse() * f2_;

        // taus according to static Euler eq
        tau1_ = -(quat_bl1.inverse()*(pbl1_- pbc_)).cross(f1_);  // torque at C due to f1
        tau2_ = -(quat_bl2.inverse()*(pbl2_- pbc_)).cross(f2_);  // torque at C due to f2

        buildDataVectors();
    }

    void setOdomData() override
    {
        prev_pose_curr_ << 0.5*ACC*pow(3*DT,2), 0, 0, 0, 0, 0, 1;
    }
};

class Cst2KFXaxisRotationPureTorque : public Cst2KFZeroMotion
{
public:
    void SetUp() override
    {
        Cst2KFZeroMotion::SetUp();
    }

    void initializeData1() override
    {
        setZeroMotionData();

        // taus according to static Euler eq
        tau1_ << M_PI/3, 0, 0; // with Identity rotational inertial matrix, rotation of PI/2 rad.s-2
        w_meas_ << M_PI/3, 0, 0;

        buildDataVectors();
    }

    void setOdomData() override
    {
        prev_pose_curr_ << 0.5*ACC*pow(3*DT,2), 0, 0, 0, 0, 0, 1;
    }
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
/////////////////////////////   TESTS   //////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

TEST_F(Cst2KFZeroMotion, ZeroMotionZeroBias)
{
    // // TEST FIRST PURE INTEGRATION
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // // COM position on Z axis is not observable with this problem
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // // COM position on Z axis is not observable with this problem
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // THEN SOLVE
    problem_->perturb();

    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;

    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState().head(2), ZERO3.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState().head(2), ZERO3.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    ASSERT_MATRIX_APPROX(KF1_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), ZERO3.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), ZERO3.head(2), 1e-6);
}

TEST_F(Cst2KFZeroMotionBias, ZeroMotionBias)
{
    problem_->perturb();

    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;
    // problem_->print(4,true,true,true);
    // std::cout << report << std::endl;

    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState().head(2), ZERO3.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState().head(2), ZERO3.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // Bias value on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), BIAS_PBC_SMAL.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), BIAS_PBC_SMAL.head(2), 1e-6);
}

TEST_F(Cst2KFXaxisLinearMotion, XLinearMotionZeroBias)
{

    Vector3d posi_diff = ZERO3;
    posi_diff[0] = 0.5*ACC * pow(3*DT,2);
    Vector3d vel_diff = ZERO3;
    vel_diff[0] = ACC * 3*DT;
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // // COM position on Z axis is not observable with this problem
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), posi_diff, 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), vel_diff, 1e-6);
    // // COM position on Z axis is not observable with this problem
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState(), posi_diff, 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), vel_diff, 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    problem_->perturb();

    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;
    // problem_->print(4, true, true, true);
    // std::cout << report << std::endl;

    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState().head(2), ZERO3.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), posi_diff, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), vel_diff, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState().head(2), posi_diff.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), vel_diff, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // Bias value on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), ZERO3.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), ZERO3.head(2), 1e-6);
}

TEST_F(Cst2KFXaxisLinearMotionPbc, XLinearMotionBias)
{
    Vector3d posi_diff = ZERO3;
    posi_diff[0] = 0.5*ACC * pow(3*DT,2);
    Vector3d vel_diff = ZERO3;
    vel_diff[0] = ACC * 3*DT;
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // // COM position on Z axis is not observable with this problem
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState().head(2), (PBCX + PBCY + PBCZ).head(2), 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    // ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), posi_diff, 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), vel_diff, 1e-6);
    // // COM position on Z axis is not observable with this problem
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState().head(2), (PBCX + PBCY + PBCZ).head(2) + posi_diff.head(2), 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), vel_diff, 1e-6);
    // ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    problem_->perturb();

    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;
    // std::cout << report << std::endl;

    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState().head(2), (PBCX + PBCY + PBCZ).head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), posi_diff, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), vel_diff, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState().head(2), (PBCX + PBCY + PBCZ).head(2) + posi_diff.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), vel_diff, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // Bias value on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), ZERO3.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), ZERO3.head(2), 1e-6);
}

TEST_F(Cst2KFXaxisLinearMotionPbcBias, XLinearMotionBias)
{
    problem_->perturb();

    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;
    // problem_->print(4, true, true, true);
    // std::cout << report << std::endl;

    Vector3d posi_diff = ZERO3;
    posi_diff[0] = 0.5*ACC * pow(3*DT,2);
    Vector3d vel_diff = ZERO3;
    vel_diff[0] = ACC * 3*DT;
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState().head(2), (PBCX + PBCY + PBCZ).head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), posi_diff, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), vel_diff, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState().head(2), (PBCX + PBCY + PBCZ).head(2) + posi_diff.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), vel_diff, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // Bias value on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), BIAS_PBC_SMAL.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), BIAS_PBC_SMAL.head(2), 1e-6);
}

TEST_F(Cst2KFXaxisLinearMotionPbcBiasPQbl, XLinearMotionBias)
{
    problem_->perturb();

    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;
    // problem_->print(4, true, true, true);
    // std::cout << report << std::endl;

    Vector3d posi_diff = ZERO3;
    posi_diff[0] = 0.5*ACC * pow(3*DT,2);
    Vector3d vel_diff = ZERO3;
    vel_diff[0] = ACC * 3*DT;
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState().head(2), (PBCX + PBCY + PBCZ).head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('P')->getState(), posi_diff, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('V')->getState(), vel_diff, 1e-6);
    // COM position on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('C')->getState().head(2), (PBCX + PBCY + PBCZ).head(2) + posi_diff.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('D')->getState(), vel_diff, 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    // Bias value on Z axis is not observable with this problem
    ASSERT_MATRIX_APPROX(KF1_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), BIAS_PBC_SMAL.head(2), 1e-6);
    ASSERT_MATRIX_APPROX(KF2_->getCaptureOf(sen_ikin_)->getSensorIntrinsic()->getState().head(2), BIAS_PBC_SMAL.head(2), 1e-6);
}


// TEST_F(Cst2KFXaxisRotationPureTorque, RotationOnlyNoSolve)
// {

//     Vector3d Lc_diff; Lc_diff << M_PI, 0, 0;
//     Vector4d q_diff; q_diff << 1,0,0,0;
//     ASSERT_MATRIX_APPROX(KF1_->getStateBlock('P')->getState(), ZERO3, 1e-6);
//     ASSERT_MATRIX_APPROX(KF1_->getStateBlock('V')->getState(), ZERO3, 1e-6);
//     ASSERT_MATRIX_APPROX(KF1_->getStateBlock('C')->getState(), ZERO3, 1e-6);
//     ASSERT_MATRIX_APPROX(KF1_->getStateBlock('D')->getState(), ZERO3, 1e-6);
//     ASSERT_MATRIX_APPROX(KF1_->getStateBlock('L')->getState(), ZERO3, 1e-6);

//     ASSERT_MATRIX_APPROX(KF2_->getStateBlock('O')->getState(), q_diff, 1e-6);
//     ASSERT_MATRIX_APPROX(KF2_->getStateBlock('L')->getState(), Lc_diff, 1e-6);
// }


int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);

    return RUN_ALL_TESTS();
}
